
.bd-main .bd-content .bd-article-container{
  max-width:100%;
  flex-grow: 1;
}

html[data-theme=light]{
  --pst-color-primary: rgb(52, 54, 99);
  --pst-color-secondary: rgb(107, 161, 174);
  --pst-color-link: rgb(74, 105, 145);
  --pst-color-inline-code: rgb(96, 141, 130);
}

:root {
  --pst-font-size-h1: 38px;
  --pst-font-size-h2: 32px;
  --pst-font-size-h3: 27px;
  --pst-font-size-h4: 22px;
  --pst-font-size-h5: 18px;
  --pst-font-size-h6: 15px;
}